 /*
  * COPYRIGHT (c) International Business Machines Corp. 2001-2017
  *
  * This program is provided under the terms of the Common Public License,
  * version 1.0 (CPL-1.0). Any use, reproduction or distribution for this
  * software constitutes recipient's acceptance of CPL-1.0 terms which can be
  * found in the file LICENSE file or at
  * https://opensource.org/licenses/cpl1.0.php
  */

#ifndef _STDLL_KEYS_H
#define _STDLL_KEYS_H

#include <pkcs11o.h>
#include <bsafe.h>

/* This header file will be eliminated when the objects are actually created
 * by the library rather than being hard coded into it. */

B_KEY_OBJ                 BSafe_Key_Object;

#endif
